# Copyright (c) 2018 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Implementation of decorators for label functions.

This module implements the label, classlabel, and staticlabel decorator
functions which can be used to decorate label functions which are part
of a FlowProject class definition.
"""


class label:
    """Decorate a :class:`~.FlowProject` class function as a label function.

    For example:

    .. code-block:: python

        class MyProject(FlowProject):

            @label()
            def foo(self, job):
                return True
    """

    def __init__(self, name=None):
        self.name = name

    def __call__(self, func):
        """Add the function as a label.

        This call operator allows the class to be used as a decorator.

        Parameters
        ----------
        func : callable
            The function to decorate.

        Returns
        -------
        callable
            The decorated function.

        """
        func._label = True
        if self.name is not None:
            func._label_name = self.name
        return func


class staticlabel(label):
    """A label decorator for staticmethods.

    This decorator implies "staticmethod"!
    """

    def __call__(self, func):
        """Add the function as a label.

        This call operator allows the class to be used as a decorator.

        Parameters
        ----------
        func : callable
            The function to decorate.

        Returns
        -------
        callable
            The decorated function.

        """
        return staticmethod(super().__call__(func))


class classlabel(label):
    """A label decorator for classmethods.

    This decorator implies "classmethod"!
    """

    def __call__(self, func):
        """Add the function as a label.

        This call operator allows the class to be used as a decorator.

        Parameters
        ----------
        func : callable
            The function to decorate.

        Returns
        -------
        callable
            The decorated function.

        """
        return classmethod(super().__call__(func))


def _is_label_func(func):
    return getattr(getattr(func, "__func__", func), "_label", False)



# Copyright (c) 2018 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Contains logic for working with flow related information stored in the signac config."""
from signac.common import config

from ..errors import ConfigKeyError

# Monkeypatch the signac config spec to include flow-specific fields.
config.cfg += """
[flow]
import_packaged_environments = boolean()
status_performance_warn_threshold = float(default=0.2)
show_traceback = boolean()
eligible_jobs_max_lines = int(default=10)
status_parallelization = string(default='none')
"""


class _GetConfigValueNoneType:
    pass


_GET_CONFIG_VALUE_NONE = _GetConfigValueNoneType()


def require_config_value(key, ns=None, default=_GET_CONFIG_VALUE_NONE):
    """Request a value from the user's configuration, failing if not available.

    Parameters
    ----------
    key : str
        The environment specific configuration key.
    ns : str
        The namespace in which to look for the key. (Default value = None)
    default
        A default value in case the key cannot be found
        within the user's configuration.

    Returns
    -------
    object
        The value or default value.

    Raises
    ------
    :class:`~.ConfigKeyError`
        If the key is not in the user's configuration
        and no default value is provided.

    """
    try:
        if ns is None:
            return config.load_config()["flow"][key]
        else:
            return config.load_config()["flow"][ns][key]
    except KeyError:
        if default is _GET_CONFIG_VALUE_NONE:
            k = str(key) if ns is None else f"{ns}.{key}"
            raise ConfigKeyError("flow." + k)
        else:
            return default


def get_config_value(key, ns=None, default=None):
    """Request a value from the user's configuration.

    Parameters
    ----------
    key : str
        The configuration key.
    ns : str
        The namespace in which to look for the key. (Default value = None)
    default
        A default value returned if the key cannot be found within the user
        configuration.

    Returns
    -------
    object
        The value if found, None if not found.

    """
    return require_config_value(key=key, ns=ns, default=default)



# Copyright (c) 2018 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Implementation of the scheduling system for the bundled ``simple-scheduler`` script."""
import json
import os
import subprocess

from .base import ClusterJob, JobStatus, Scheduler, _call_submit


class SimpleScheduler(Scheduler):
    """Implementation of the abstract Scheduler class for the bundled ``simple-scheduler``.

    The package signac-flow includes a script in ``bin/simple-scheduler`` that
    is a simple model of a cluster job scheduler. The ``simple-scheduler``
    script is designed primarily for testing and demonstration.

    This class can submit cluster jobs to the built-in simple scheduler and
    query their current status.
    """

    @classmethod
    def is_present(cls):
        """Return True if a SimpleScheduler is detected."""
        return bool(os.environ.get("SIMPLE_SCHEDULER"))

    def __init__(self):
        self.cmd = os.environ["SIMPLE_SCHEDULER"].split()
        self.submit_cmd = self.cmd + ["submit"]

    def jobs(self):
        """Yield cluster jobs by querying the scheduler."""
        cmd = self.cmd + ["status", "--json"]
        status = json.loads(subprocess.check_output(cmd).decode("utf-8"))
        for _id, doc in status.items():
            yield ClusterJob(doc["job_name"], JobStatus(doc["status"]))

    def submit(self, script, *, pretend=False, **kwargs):
        r"""Submit a job script for execution to the scheduler.

        Parameters
        ----------
        script : str
            The job script submitted for execution.
        pretend : bool
            If True, do not actually submit the script, but only simulate the
            submission. Can be used to test whether the submission would be
            successful. A successful "pretend" submission is not
            guaranteed to succeed. (Default value = False)
        \*\*kwargs
            Keyword arguments (ignored).

        Returns
        -------
        bool
            True if the submission command succeeds (or in pretend mode).

        Raises
        ------
        :class:`~flow.errors.SubmitError`
            If the submission command fails.


        """
        return _call_submit(self.submit_cmd, script, pretend)



# Copyright (c) 2020 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Environments for the Minnesota Supercomputing Institute at University of Minnesota."""
from ..environment import DefaultSlurmEnvironment


class MangiEnvironment(DefaultSlurmEnvironment):
    """Environment profile for the Mangi supercomputer.

    Mangi is managed by the Minnesota Supercomputing Institute (University of
    Minnesota).

    Cluster information: https://www.msi.umn.edu/mangi

    Queue information: https://www.msi.umn.edu/queues
    """

    hostname_pattern = r".*\.mangi\.msi\.umn\.edu"
    template = "slurm.sh"
    cores_per_node = 1


__all__ = ["MangiEnvironment"]



# Copyright (c) 2020 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Aggregation allows the definition of operations on multiple jobs.

Operations are each associated with an aggregator that determines how
jobs are grouped before being passed as arguments to the operation. The
default aggregator produces individual jobs.
"""
import itertools
from collections.abc import Iterable, Mapping
from hashlib import md5


def _get_unique_function_id(func):
    """Generate unique id for the provided function.

    Hashing the bytecode rather than directly hashing the function allows for
    the comparison of internal functions like ``self._aggregator_function``
    or ``self._select`` that may have the same definitions but different
    hashes simply because they are distinct objects.

    It is possible for equivalent functions to have different ids if the
    bytecode is not identical.

    Parameters
    ----------
    func : callable
        The function to be hashed.

    Returns
    -------
    str
        The hash of the function's bytecode if possible, otherwise the hash
        of the function.

    """
    try:
        return hash(func.__code__.co_code)
    except AttributeError:  # Cannot access function's compiled bytecode
        return hash(func)


class _aggregator:
    """Decorator for operation functions that operate on aggregates.

    By default, if the ``aggregator_function`` is ``None``, an aggregate of all
    jobs will be created.

    Examples
    --------
    The code block below defines a :class:`~.FlowOperation` that prints the
    total length of the provided aggregate of jobs.

    .. code-block:: python

        @aggregator()
        @FlowProject.operation
        def foo(*jobs):
            print(len(jobs))

    Parameters
    ----------
    aggregator_function : callable or None
        A callable that performs aggregation of jobs. It takes in a list of
        jobs and can return or yield subsets of jobs as an iterable. The
        default behavior is creating a single aggregate of all jobs.
    sort_by : str, callable, or None
        Before aggregating, sort the jobs by a given statepoint parameter. If
        the argument is callable, this will be passed as the callable
        argument to :func:`sorted`. If None, no sorted is performed (Default
        value = None).
    sort_ascending : bool
        True if the jobs are to be sorted in ascending order. (Default value
        = True)
    select : callable or None
        Condition for filtering individual jobs. This is passed as the
        callable argument to :func:`filter`. If None, no filtering is
        performed. (Default value = None)

    """

    def __init__(
        self, aggregator_function=None, sort_by=None, sort_ascending=True, select=None
    ):
        if aggregator_function is None:

            def aggregator_function(jobs):
                yield tuple(jobs) if jobs else ()

        if not callable(aggregator_function):
            raise TypeError(
                "Expected aggregator_function to be callable, got "
                f"{type(aggregator_function)}"
            )
        if sort_by is not None and not (isinstance(sort_by, str) or callable(sort_by)):
            raise TypeError(
                f"Expected sort_by parameter to be str or callable, got {type(sort_by)}"
            )
        if select is not None and not callable(select):
            raise TypeError(
                f"Expected select parameter to be callable, got {type(select)}"
            )

        # Set the ``_is_default_aggregator`` attribute to False by default. But if
        # the "non-aggregate" aggregator object i.e. aggregator.groupsof(1) is
        # created using the class method, then we explicitly set the
        # ``_is_default_aggregator`` attribute to True.
        self._is_default_aggregator = False
        self._aggregator_function = aggregator_function
        self._sort_by = sort_by
        self._sort_ascending = bool(sort_ascending)
        self._select = select

    @classmethod
    def groupsof(cls, num=1, sort_by=None, sort_ascending=True, select=None):
        """Aggregate jobs into groupings of a given size.

        By default, creates aggregates consisting of a single job.

        If the number of jobs present in the project is not divisible by the
        number provided by the user, the last aggregate will be smaller and
        contain the remaining jobs. For instance, if 10 jobs are present in a
        project and they are aggregated in groups of 3, then the generated
        aggregates will have lengths 3, 3, 3, and 1.

        Examples
        --------
        The code block below shows how to aggregate jobs in groups of 2.

        .. code-block:: python

            @aggregator.groupsof(num=2)
            @FlowProject.operation
            def foo(*jobs):
                print(len(jobs))

        Parameters
        ----------
        num : int
            The default size of aggregates excluding the final aggregate.
        sort_by : str or None
            Before aggregating, sort the jobs by a given statepoint parameter.
            The default behavior is no sorting.
        sort_ascending : bool
            States if the jobs are to be sorted in ascending order.
            The default value is True.
        select : callable or None
            Condition for filtering individual jobs. This is passed as the
            callable argument to :func:`filter`. The default behavior is no
            filtering.

        Returns
        -------
        aggregator : :class:`~.aggregator`
            The :meth:`~.groupsof` aggregator.

        """
        try:
            if num != int(num):
                raise ValueError("The num parameter should be an integer")
            num = int(num)
            if num <= 0:
                raise ValueError("The num parameter should have a value greater than 0")
        except TypeError:
            raise TypeError("The num parameter should be an integer")

        # This method is similar to the `grouper` method documented here:
        # https://docs.python.org/3/library/itertools.html#itertools.zip_longest
        # However, this function does not have a fill value.
        # Source of this implementation: https://stackoverflow.com/a/31185097
        def aggregator_function(jobs):
            iterable = iter(jobs)
            return iter(lambda: tuple(itertools.islice(iterable, num)), tuple())

        aggregator_instance = cls(aggregator_function, sort_by, sort_ascending, select)

        if num == 1 and sort_by is None and select is None and sort_ascending:
            aggregator_instance._is_default_aggregator = True

        return aggregator_instance

    @classmethod
    def groupby(cls, key, default=None, sort_by=None, sort_ascending=True, select=None):
        """Aggregate jobs according to matching state point values.

        Examples
        --------
        The code block below provides an example of how to aggregate jobs
        having a common state point parameter ``'sp'`` whose value, if not
        found, is replaced by a default value of -1.

        .. code-block:: python

            @aggregator.groupby(key='sp', default=-1)
            @FlowProject.operation
            def foo(*jobs):
                print(len(jobs))

        Parameters
        ----------
        key : str, Iterable[str], or callable
            The method by which jobs are grouped. It may be a state point key
            or an iterable of state point keys whose values define the
            groupings. It may also be an arbitrary callable of
            :class:`~signac.contrib.job.Job` when greater flexibility is
            needed.
        default : Any
            Default value used for grouping if the key is missing or invalid.
            If ``key`` is an iterable, the default value must be a sequence
            of equal length. If ``key`` is a callable, this argument is
            ignored. If None, the provided keys must exist for all jobs.
            (Default value = None)
        sort_by : str or None
            State point parameter used to sort the jobs before grouping.
            The default value is None, which does no sorting.
        sort_ascending : bool
            Whether jobs are to be sorted in ascending order. (Default value
            = True)
        select : callable or None
            Condition for filtering individual jobs. This is passed as the
            callable argument to :func:`filter`. The default behavior is no
            filtering.

        Returns
        -------
        aggregator : :class:`~.aggregator`
            The :meth:`~.groupby` aggregator.

        """
        if isinstance(key, str):
            if default is None:

                def keyfunction(job):
                    return job.statepoint[key]

            else:

                def keyfunction(job):
                    return job.statepoint.get(key, default)

        elif isinstance(key, Iterable):
            keys = list(key)
            if default is None:

                def keyfunction(job):
                    return [job.statepoint[key] for key in keys]

            else:
                if isinstance(default, Iterable):
                    if len(default) != len(keys):
                        raise ValueError(
                            "Expected length of default argument is "
                            f"{len(keys)}, got {len(default)}."
                        )
                else:
                    raise TypeError(
                        "Invalid default argument. Expected Iterable, "
                        f"got {type(default)}"
                    )

                def keyfunction(job):
                    return [
                        job.statepoint.get(key, default_value)
                        for key, default_value in zip(keys, default)
                    ]

        elif callable(key):
            keyfunction = key
        else:
            raise TypeError(
                "Invalid key argument. Expected str, Iterable, "
                f"or a callable, got {type(key)}"
            )

        def aggregator_function(jobs):
            for key, group in itertools.groupby(
                sorted(jobs, key=keyfunction), key=keyfunction
            ):
                yield tuple(group)

        return cls(aggregator_function, sort_by, sort_ascending, select)

    def __eq__(self, other):
        """Test equality with another aggregator."""
        if not isinstance(other, type(self)):
            return NotImplemented
        # It is not possible to compare aggregators, even with equivalent
        # aggregator functions. Moreover, the code objects created by
        # _get_unique_function_id do not account for differences in the bound
        # parameters. Thus, the only meaningful comparison is whether both
        # aggregators are the default aggregator (and thus equivalent).
        return self._is_default_aggregator and other._is_default_aggregator

    def __hash__(self):
        """Hash this aggregator."""
        return hash(
            (
                self._is_default_aggregator,
                self._sort_ascending,
                _get_unique_function_id(self._sort_by),
                _get_unique_function_id(self._aggregator_function),
                _get_unique_function_id(self._select),
            )
        )

    def _create_AggregateStore(self, project):
        """Create the actual collections of jobs to be sent to aggregate operations.

        The :class:`aggregator` class is just a decorator that provides a
        signal for operation functions that should be treated as aggregate
        operations and information on how to perform the aggregation. This
        function generates the classes that actually hold the aggregates
        (tuples of jobs) to which aggregate operations will be applied.

        Parameters
        ----------
        project : :class:`signac.contrib.project.Project`
            A signac project used to fetch jobs for creating aggregates.

        Returns
        -------
        :class:`~._BaseAggregateStore`
            The aggregate store.

        """
        if self._is_default_aggregator:
            return _DefaultAggregateStore(project)
        else:
            return _AggregateStore(self, project)

    def __call__(self, func=None):
        """Add this aggregator to a provided operation.

        This call operator allows the class to be used as a decorator.

        Parameters
        ----------
        func : callable
            The function to decorate.

        """
        if callable(func):
            setattr(func, "_flow_aggregate", self)
            return func
        else:
            raise TypeError(
                "Invalid argument passed while calling the aggregate "
                f"instance. Expected a callable, got {type(func)}."
            )


class _BaseAggregateStore(Mapping):
    """Base abstract class for aggregate stores.

    An aggregate store is a mapping from aggregate ids to aggregates, where
    an aggregate is defined as a tuple of instances of
    :class:`signac.contrib.job.Job`.
    """

    def __init__(self, project):
        self._project = project

    def __iter__(self):
        yield from self.keys()


class _AggregateStore(_BaseAggregateStore):
    """Class containing all aggregates associated with an :class:`aggregator`.

    Iterating over this object yields aggregate ids, which can be used as
    indices to return the corresponding aggregates.

    Parameters
    ----------
    aggregator : :class:`aggregator`
        aggregator object used to generate aggregates for this store.
    project : :class:`flow.FlowProject` or :class:`signac.contrib.project.Project`
        A signac project containing the jobs that will be used to create
        aggregates.

    """

    def __init__(self, aggregator, project):
        self._aggregator = aggregator
        super().__init__(project)

        # We need to register the aggregates for this instance using the
        # provided project. After registering, we store the aggregates mapped
        # with the ids using :func:`get_aggregate_id`.
        self._aggregates_by_id = {}
        self._register_aggregates()

    def __getitem__(self, id):
        """Get the aggregate corresponding to the provided id."""
        try:
            return self._aggregates_by_id[id]
        except KeyError:
            raise KeyError(f"Aggregate id {id} could not be found.")

    def __contains__(self, id):
        """Return whether this instance contains an aggregate (by aggregate id).

        Parameters
        ----------
        id : str
            The id of an aggregate of jobs.

        Returns
        -------
        bool
            Whether this instance contains the aggregate.

        """
        return id in self._aggregates_by_id

    def __len__(self):
        return len(self._aggregates_by_id)

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return NotImplemented
        return self._aggregator == other._aggregator

    def __hash__(self):
        return hash(self._aggregator)

    def keys(self):
        return self._aggregates_by_id.keys()

    def values(self):
        return self._aggregates_by_id.values()

    def items(self):
        return self._aggregates_by_id.items()

    def _register_aggregates(self):
        """Register aggregates from the project.

        This is called at instantiation to generate and store aggregates.

        Every aggregate is required to be a tuple of jobs.
        """
        for aggregate in self._generate_aggregates():
            for job in aggregate:
                if job not in self._project:
                    raise LookupError(
                        f"The signac job {job.get_id()} not found in {self._project}"
                    )
            try:
                stored_aggregate = tuple(aggregate)
            except TypeError:  # aggregate is not iterable
                raise ValueError("Invalid aggregator_function provided by the user.")
            # Store aggregate by id to allow searching by id
            self._aggregates_by_id[
                _get_aggregate_id(stored_aggregate)
            ] = stored_aggregate

    def _generate_aggregates(self):
        jobs = self._project
        if self._aggregator._select is not None:
            jobs = filter(self._aggregator._select, jobs)
        if self._aggregator._sort_by is None:
            jobs = list(jobs)
        else:
            if callable(self._aggregator._sort_by):
                sort_function = self._aggregator._sort_by
            else:

                def sort_function(job):
                    return job.statepoint[self._aggregator._sort_by]

            jobs = sorted(
                jobs,
                key=sort_function,
                reverse=not self._aggregator._sort_ascending,
            )
        yield from self._aggregator._aggregator_function(jobs)


class _DefaultAggregateStore(_BaseAggregateStore):
    """Aggregate storage wrapper for the default aggregator.

    This class holds the information of the project associated with an
    operation function using the default aggregator, i.e.
    ``aggregator.groupsof(1)``.

    Iterating over this object yields tuples each containing one job from the project.

    Parameters
    ----------
    project : :class:`flow.FlowProject` or :class:`signac.contrib.project.Project`
        A signac project used to fetch jobs for creating aggregates.

    """

    def __init__(self, project):
        super().__init__(project)
        # Below, we store repr(project), which defines the hash and equality
        # operators of this class. This class must be hashable because it is
        # used as a dict key. However, when unpickling a FlowProject, this
        # object's hash must be computed *before* the FlowProject is fully
        # initialized. Thus, it is not possible to execute repr(project) when
        # hashing the instance at the time of unpickling. This means that this
        # class cannot be unpickled unless we pre-emptively compute and store
        # the repr.
        self._project_repr = repr(project)

    def __getitem__(self, id):
        """Return an aggregate of one job from its job id.

        Parameters
        ----------
        id : str
            The job id.

        """
        try:
            return (self._project.open_job(id=id),)
        except KeyError:
            raise KeyError(f"Aggregate id {id} could not be found.")

    def __contains__(self, id):
        """Return whether this instance contains a job (by job id).

        Parameters
        ----------
        id : str
            The job id.

        """
        try:
            self._project.open_job(id=id)
        except KeyError:
            return False
        except LookupError:
            raise
        else:
            return True

    def __len__(self):
        return len(self._project)

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return NotImplemented
        return self._project_repr == other._project_repr

    def __hash__(self):
        return hash(self._project_repr)

    def keys(self):
        for job in self._project:
            yield job.get_id()

    def values(self):
        for job in self._project:
            yield (job,)

    def items(self):
        for job in self._project:
            yield (job.get_id(), (job,))


def _get_aggregate_id(aggregate):
    """Generate aggregate id for an aggregate of jobs.

    The aggregate id is a unique hash identifying a tuple of jobs. The
    aggregate id is sensitive to the order of the jobs in the aggregate. The
    id of an aggregate containing one job is that job's id (the hash of its
    state point).

    Parameters
    ----------
    aggregate : tuple of :class:`~signac.contrib.job.Job`
        Aggregate of signac jobs.

    Returns
    -------
    str
        The generated aggregate id.

    """
    if len(aggregate) == 1:
        # Return job id as it's already unique
        return aggregate[0].get_id()

    id_string = ",".join(job.get_id() for job in aggregate)
    hash_ = md5(id_string.encode("utf-8")).hexdigest()
    return f"agg-{hash_}"



# Copyright (c) 2017 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Environments for XSEDE supercomputers."""
import logging
import os

from ..environment import DefaultSlurmEnvironment, template_filter

logger = logging.getLogger(__name__)


class CometEnvironment(DefaultSlurmEnvironment):
    """Environment profile for the Comet supercomputer.

    https://www.sdsc.edu/services/hpc/hpc_systems.html#comet
    """

    hostname_pattern = "comet"
    template = "comet.sh"
    cores_per_node = 24
    mpi_cmd = "ibrun"

    @classmethod
    def add_args(cls, parser):
        """Add arguments to parser.

        Parameters
        ----------
        parser : :class:`argparse.ArgumentParser`
            The argument parser where arguments will be added.

        """
        super().add_args(parser)

        parser.add_argument(
            "--partition",
            choices=["compute", "gpu", "gpu-shared", "shared", "large-shared", "debug"],
            default="shared",
            help="Specify the partition to submit to.",
        )

        parser.add_argument(
            "--job-output",
            help=(
                "What to name the job output file. "
                "If omitted, uses the system default "
                '(slurm default is "slurm-%%j.out").'
            ),
        )


_STAMPEDE_OFFSET = os.environ.get("_FLOW_STAMPEDE_OFFSET_`", 0)


class Stampede2Environment(DefaultSlurmEnvironment):
    """Environment profile for the Stampede2 supercomputer.

    https://www.tacc.utexas.edu/systems/stampede2
    """

    hostname_pattern = ".*stampede2"
    template = "stampede2.sh"
    cores_per_node = 48
    mpi_cmd = "ibrun"
    offset_counter = 0
    base_offset = _STAMPEDE_OFFSET

    @template_filter
    def return_and_increment(cls, increment):
        """Increment the base offset, then return the prior value.

        Note that this filter is designed for use at submission time, and the
        environment variable will be used upon script generation. At run time,
        the base offset will be set only once (when run initializes the
        environment).

        Parameters
        ----------
        increment : int
            The increment to apply to the base offset.

        Returns
        -------
        int
            The value of the base offset before incrementing.

        """
        cls.base_offset += increment
        return cls.base_offset - increment

    @template_filter
    def decrement_offset(cls, decrement):
        """Decrement the offset value.

        This function is a hackish solution to get around the fact that Jinja
        has very limited support for direct modification of Python objects, and
        we need to be able to reset the offset between bundles.

        Parameters
        ----------
        decrement : int
            The decrement to apply to the base offset.

        Returns
        -------
        str
            Empty string (to render nothing in the jinja template).

        """
        cls.base_offset -= int(decrement)
        return ""

    @classmethod
    def add_args(cls, parser):
        """Add arguments to parser.

        Parameters
        ----------
        parser : :class:`argparse.ArgumentParser`
            The argument parser where arguments will be added.

        """
        super().add_args(parser)
        parser.add_argument(
            "--partition",
            choices=[
                "development",
                "normal",
                "large",
                "flat-quadrant",
                "skx-dev",
                "skx-normal",
                "skx-large",
                "long",
            ],
            default="skx-normal",
            help="Specify the partition to submit to.",
        )
        parser.add_argument(
            "--job-output",
            help=(
                "What to name the job output file. "
                "If omitted, uses the system default "
                '(slurm default is "slurm-%%j.out").'
            ),
        )

    @classmethod
    def _get_mpi_prefix(cls, operation, parallel):
        """Get the MPI prefix based on directives and an offset.

        Parameters
        ----------
        operation : :class:`flow.project._JobOperation`
            The operation to be prefixed.
        parallel : bool
            If True, operations are assumed to be executed in parallel, which
            means that the number of total tasks is the sum of all tasks
            instead of the maximum number of tasks. Default is set to False.

        Returns
        -------
        str
            The prefix to be added to the operation's command.

        """
        if operation.directives.get("nranks"):
            prefix = "{} -n {} -o {} task_affinity ".format(
                cls.mpi_cmd,
                operation.directives["nranks"],
                cls.base_offset + cls.offset_counter,
            )
            if parallel:
                cls.offset_counter += operation.directives["nranks"]
        else:
            prefix = ""
        return prefix


class Bridges2Environment(DefaultSlurmEnvironment):
    """Environment profile for the Bridges-2 supercomputer.

    https://www.psc.edu/resources/bridges-2/user-guide
    """

    hostname_pattern = r".*\.bridges2\.psc\.edu$"
    template = "bridges2.sh"
    cores_per_node = 128
    mpi_cmd = "mpirun"

    @classmethod
    def add_args(cls, parser):
        """Add arguments to parser.

        Parameters
        ----------
        parser : :class:`argparse.ArgumentParser`
            The argument parser where arguments will be added.

        """
        super().add_args(parser)
        parser.add_argument(
            "--partition",
            choices=[
                "RM",
                "RM-shared",
                "RM-small",
                "EM",
                "GPU",
                "GPU-shared",
            ],
            default="RM-shared",
            help="Specify the partition to submit to.",
        )


__all__ = [
    "CometEnvironment",
    "Stampede2Environment",
    "Bridges2Environment",
]



# Copyright (c) 2018 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Defines operation decorators and a simple command line interface ``run``.

This module implements the run() function, which when called equips a regular
Python module with a command line interface. This interface can be used to
execute functions defined within the same module that operate on a signac data
space.

See also: :class:`~.FlowProject`.
"""
import inspect
import logging
from functools import wraps
from textwrap import indent

from .environment import ComputeEnvironment

logger = logging.getLogger(__name__)


def cmd(func):
    """Indicate that ``func`` returns a shell command with this decorator.

    If this function is an operation function defined by :class:`~.FlowProject`, it will
    be interpreted to return a shell command, instead of executing the function itself.

    For example:

    .. code-block:: python

        @FlowProject.operation
        @flow.cmd
        def hello(job):
            return "echo {job.id}"

    .. note::
        The final shell command generated for :meth:`~.FlowProject.run` or
        :meth:`~.FlowProject.submit` still respects directives and will prepend e.g. MPI or OpenMP
        prefixes to the shell command provided here.
    """
    if getattr(func, "_flow_with_job", False):
        raise RuntimeError(
            "@cmd should appear below the @with_job decorator in your script"
        )
    setattr(func, "_flow_cmd", True)
    return func


def with_job(func):
    """Use ``arg`` as a context manager for ``func(arg)`` with this decorator.

    If this function is an operation function defined by :class:`~.FlowProject`, it will
    be the same as using ``with job:``.

    For example:

    .. code-block:: python

        @FlowProject.operation
        @flow.with_job
        def hello(job):
            print("hello {}".format(job))

    Is equivalent to:

    .. code-block:: python

        @FlowProject.operation
        def hello(job):
            with job:
                print("hello {}".format(job))

    This also works with the `@cmd` decorator:

    .. code-block:: python

        @FlowProject.operation
        @with_job
        @cmd
        def hello(job):
            return "echo 'hello {}'".format(job)

    Is equivalent to:

    .. code-block:: python

        @FlowProject.operation
        @cmd
        def hello_cmd(job):
            return 'trap "cd `pwd`" EXIT && cd {} && echo "hello {job}"'.format(job.ws)
    """

    @wraps(func)
    def decorated(job):
        with job:
            if getattr(func, "_flow_cmd", False):
                return f'trap "cd $(pwd)" EXIT && cd {job.ws} && {func(job)}'
            else:
                return func(job)

    setattr(decorated, "_flow_with_job", True)
    return decorated


class directives:
    """Decorator for operation functions to provide additional execution directives.

    Directives can for example be used to provide information about required resources
    such as the number of processes required for execution of parallelized operations.
    For more information, read about :ref:`signac-docs:directives`.

    """

    def __init__(self, **kwargs):
        self.kwargs = kwargs

    @classmethod
    def copy_from(cls, func):
        """Copy directives from another operation."""
        return cls(**getattr(func, "_flow_directives", {}))

    def __call__(self, func):
        """Add directives to the function.

        This call operator allows the class to be used as a decorator.

        Parameters
        ----------
        func : callable
            The function to decorate.

        Returns
        -------
        callable
            The decorated function.

        """
        directives = getattr(func, "_flow_directives", {})
        directives.update(self.kwargs)
        setattr(func, "_flow_directives", directives)
        return func


def _document_directive(directive):
    name = directive._name
    name = name.replace("_", r"\_")
    doc = directive.__doc__
    return f"**{name}**\n\n{doc}"


_directives_to_document = (
    ComputeEnvironment._get_default_directives()._directive_definitions.values()
)
directives.__doc__ += indent(
    "\n**Supported Directives:**\n\n"
    + "\n\n".join(
        _document_directive(directive) for directive in _directives_to_document
    ),
    "    ",
)


def _get_operations(include_private=False):
    """Yield the name of all functions that qualify as an operation function.

    The module is inspected and all functions that have only one argument
    is yielded. Unless the 'include_private' argument is True, all private
    functions, that means the name starts with one or more '_' characters
    are ignored.
    """
    module = inspect.getmodule(inspect.currentframe().f_back.f_back)
    for name, obj in inspect.getmembers(module):
        if not include_private and name.startswith("_"):
            continue
        if inspect.isfunction(obj):
            signature = inspect.getfullargspec(obj)
            if len(signature.args) == 1:
                yield name


__all__ = ["cmd", "directives", "with_job"]



# Copyright (c) 2018 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Definition of base classes for the scheduling system."""
import enum
import subprocess
import tempfile
import time
from abc import ABC, abstractmethod

from ..errors import SubmitError


class JobStatus(enum.IntEnum):
    """Classifies the job's execution status."""

    unknown = 1
    """Unknown cluster job status."""

    registered = 2
    """The cluster job is registered with the scheduler, but no other status is known."""

    inactive = 3
    """The cluster job is inactive.

    This includes states like completed, cancelled, or timed out.
    """

    submitted = 4
    """The cluster job has been submitted.

    Note that this state is never returned by a scheduler, but is an assumed
    state immediately after a cluster job is submitted.
    """

    held = 5
    """The cluster job is held."""

    queued = 6
    """The cluster job is queued."""

    active = 7
    """The cluster job is actively running."""

    error = 8
    """The cluster job is in an error or failed state."""

    placeholder = 127
    """A placeholder state that is used for status rendering when no operations are eligible."""

    user = 128
    """All user-defined states must be >=128 in value."""


class ClusterJob:
    """Class representing a cluster job."""

    def __init__(self, job_id, status=None):
        self._job_id = job_id
        self._status = status

    def _id(self):
        return self._job_id

    def __str__(self):
        """Return job ID string."""
        return str(self._id())

    def name(self):
        """Return the name of the cluster job."""
        return self._id()

    def status(self):
        """Return the status of the cluster job."""
        return self._status


class Scheduler(ABC):
    """Abstract base class for schedulers."""

    # The UNIX time stamp of the last scheduler query.
    _last_query = None

    # The amount of time in seconds to wait between scheduler queries.
    # Repeated scheduler queries might risk a denial-of-service attack.
    _dos_timeout = 10

    @classmethod
    def _prevent_dos(cls):
        """Prevent denial of service by enforcing a back-off period.

        This method should *always* be called before querying the scheduler.

        This method will raise an exception if it is called more than
        once within a time window defined by the value of ``_dos_timeout``.
        This is to prevent an (accidental) denial-of-service attack on the
        scheduling system.
        """
        if cls._last_query is not None:
            if time.time() - cls._last_query < cls._dos_timeout:
                raise RuntimeError("Too many scheduler requests within a short time!")
        cls._last_query = time.time()

    @abstractmethod
    def jobs(self):
        """Yield all cluster jobs.

        Yields
        ------
        :class:`.ClusterJob`
            Cluster job.

        """
        raise NotImplementedError()

    @abstractmethod
    def submit(self, script, **kwargs):
        """Submit a job script to the scheduler for execution."""
        raise NotImplementedError()

    @classmethod
    @abstractmethod
    def is_present(cls):
        """Return True if the scheduler is detected."""
        raise NotImplementedError()


def _call_submit(submit_cmd, script, pretend):
    """Call submit command with a temporary script file.

    Parameters
    ----------
    submit_cmd : list[str]
        List of strings composing the submission command and any flags.
    script : str
        Script as a string.
    pretend : bool
        If True, the script will be printed to screen instead of submitted.

    Returns
    -------
    bool
        True if the submission command succeeds (or in pretend mode).

    Raises
    ------
    :class:`~flow.errors.SubmitError`
        If the submission command fails.

    """
    submit_cmd_string = " ".join(submit_cmd)
    if pretend:
        print(f"# Submit command: {submit_cmd_string}")
        print(script)
        print()
    else:
        with tempfile.NamedTemporaryFile() as tmp_submit_script:
            tmp_submit_script.write(str(script).encode("utf-8"))
            tmp_submit_script.flush()
            submit_cmd.append(tmp_submit_script.name)
            try:
                subprocess.check_output(
                    submit_cmd, stderr=subprocess.STDOUT, universal_newlines=True
                )
            except subprocess.CalledProcessError as error:
                raise SubmitError(
                    f"Error when calling submission command {submit_cmd_string}:\n{error.output}"
                )

    return True



# Copyright (c) 2018 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Implementation of the scheduling system for LSF schedulers.

This module implements the Scheduler and ClusterJob classes for LSF.
"""
import errno
import getpass
import json
import logging
import subprocess

from .base import ClusterJob, JobStatus, Scheduler, _call_submit

logger = logging.getLogger(__name__)


def _parse_status(s):
    if s in ["PEND", "WAIT"]:
        return JobStatus.queued
    elif s == "RUN":
        return JobStatus.active
    elif s in ["SSUSP", "USUSP", "PSUSP"]:
        return JobStatus.held
    elif s == "DONE":
        return JobStatus.inactive
    elif s == "EXIT":
        return JobStatus.error
    return JobStatus.registered


def _fetch(user=None):
    """Fetch the cluster job status information from the LSF scheduler.

    Parameters
    ----------
    user : str
        Limit the status information to cluster jobs submitted by user.
        (Default value = None)

    Yields
    ------
    :class:`~.LSFJob`
        LSF cluster job.

    """
    if user is None:
        user = getpass.getuser()

    cmd = ["bjobs", "-json", "-u", user]
    try:
        result = json.loads(subprocess.check_output(cmd).decode("utf-8"))
    except subprocess.CalledProcessError:
        raise
    except OSError as error:
        if error.errno != errno.ENOENT:
            raise
        else:
            raise RuntimeError("LSF not available.")
    except json.decoder.JSONDecodeError:
        raise RuntimeError("Could not parse LSF JSON output.")

    for record in result["RECORDS"]:
        yield LSFJob(record)


class LSFJob(ClusterJob):
    """An LSFJob is a ClusterJob managed by an LSF scheduler."""

    def __init__(self, record):
        self.record = record
        self._job_id = record["JOBID"]
        self._status = _parse_status(record["STAT"])

    def name(self):
        """Return the name of the cluster job."""
        return self.record["JOB_NAME"]


class LSFScheduler(Scheduler):
    r"""Implementation of the abstract Scheduler class for LSF schedulers.

    This class can submit cluster jobs to a LSF scheduler and query their
    current status.

    Parameters
    ----------
    user : str
        Limit the status information to cluster jobs submitted by user.
    \*\*kwargs
        Forwarded to the parent constructor.

    """

    # The standard command used to submit jobs to the LSF scheduler.
    submit_cmd = ["bsub"]

    def __init__(self, user=None):
        self.user = user

    def jobs(self):
        """Yield cluster jobs by querying the scheduler."""
        self._prevent_dos()
        yield from _fetch(user=self.user)

    def submit(
        self, script, *, after=None, hold=False, pretend=False, flags=None, **kwargs
    ):
        r"""Submit a job script for execution to the scheduler.

        Parameters
        ----------
        script : str
            The job script submitted for execution.
        after : str
            Execute the submitted script after a job with this id has
            completed. (Default value = None)
        hold : bool
            Whether to hold the job upon submission. (Default value = False)
        pretend : bool
            If True, do not actually submit the script, but only simulate the
            submission. Can be used to test whether the submission would be
            successful. Please note: A successful "pretend" submission is not
            guaranteed to succeed. (Default value = False)
        flags : list
            Additional arguments to pass through to the scheduler submission
            command. (Default value = None)
        \*\*kwargs
            Additional keyword arguments (ignored).

        Returns
        -------
        bool
            True if the submission command succeeds (or in pretend mode).

        Raises
        ------
        :class:`~flow.errors.SubmitError`
            If the submission command fails.

        """
        if flags is None:
            flags = []
        elif isinstance(flags, str):
            flags = flags.split()

        submit_cmd = self.submit_cmd + flags

        if after is not None:
            submit_cmd.extend(["-w", '"done({})"'.format(after.split(".")[0])])

        if hold:
            submit_cmd += ["-H"]

        return _call_submit(submit_cmd, script, pretend)

    @classmethod
    def is_present(cls):
        """Return True if an LSF scheduler is detected."""
        try:
            subprocess.check_output(["bjobs", "-V"], stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError:
            return True
        except OSError:
            return False
        else:
            return True



# Copyright (c) 2018 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""The main command line interface for signac-flow.

The interface is accessible via the `flow` command and allows users to
initialize FlowProject class definitions directly from the command line.

Execute `flow --help` for more information.
"""
import argparse
import logging
import sys

from signac import get_project, init_project

from . import __version__, template

logger = logging.getLogger(__name__)


def main_init(args):
    """Initialize a FlowProject from a template.

    The available templates are defined in the template module.
    """
    if not args.alias.isidentifier():
        raise ValueError(
            "The alias '{}' is not a valid Python identifier and can therefore "
            "not be used as a FlowProject alias.".format(args.alias)
        )
    try:
        get_project()
    except LookupError:
        init_project(name=args.alias)
        print(
            "Initialized signac project with name '{}' in "
            "current directory.".format(args.alias),
            file=sys.stderr,
        )
    try:
        return template.init(alias=args.alias, template=args.template)
    except OSError as error:
        raise RuntimeError(
            f"Error occurred while trying to initialize a flow project: {error}"
        )


def main():
    """Define the 'flow' command line interface.

    This function defines the main 'flow' command line interface which can be used
    to initialize FlowProject modules from different templates as well as print the
    version of the installed signac-flow package.
    """
    parser = argparse.ArgumentParser(
        description="flow provides the basic components to set up workflows for "
        "projects as part of the signac framework."
    )
    parser.add_argument(
        "--debug", action="store_true", help="Show traceback on error for debugging."
    )
    parser.add_argument(
        "--version", action="store_true", help="Display the version number and exit."
    )

    subparsers = parser.add_subparsers()

    # the flow init command
    parser_init = subparsers.add_parser(
        "init", help="Initialize a signac-flow project."
    )
    parser_init.set_defaults(func=main_init)
    parser_init.add_argument(
        "alias",
        type=str,
        nargs="?",
        default="project",
        help="Name of the flow project module to initialize. "
        "This name will also be used to initialize a signac project in case that "
        "no project was initialized prior to calling 'init'.",
    )
    parser_init.add_argument(
        "-t",
        "--template",
        type=str,
        choices=tuple(sorted(template.TEMPLATES)),
        default="minimal",
        help="Specify a specific to template to use.",
    )

    if "--version" in sys.argv:
        print("signac-flow", __version__)
        sys.exit(0)

    args = parser.parse_args()

    if args.debug:
        logging.basicConfig(level=logging.DEBUG)
    if not hasattr(args, "func"):
        parser.print_usage()
        sys.exit(2)
    try:
        args.func(args)
    except KeyboardInterrupt:
        sys.stderr.write("\n")
        sys.stderr.write("Interrupted.\n")
        if args.debug:
            raise
        sys.exit(1)
    except Exception as error:
        sys.stderr.write(f"{str(error)}\n")
        if args.debug:
            raise
        sys.exit(1)
    else:
        sys.exit(0)


if __name__ == "__main__":
    main()



"""
    Directive Syntax
    ~~~~~~~~~~~~~~~~~

    This syntax is inspired by reStructuredText. The syntax is very powerful,
    that you can define a lot of custom features by your own.

    The syntax looks like::

        .. directive-name:: directive value
           :option-key: option value
           :option-key: option value

           full featured markdown text here

    :copyright: (c) Hsiaoming Yang
"""

import re

__all__ = ['Directive']


DIRECTIVE_PATTERN = re.compile(
    r'\.\.( +)(?P<name>[a-zA-Z0-9_-]+)\:\: *(?P<value>[^\n]*)\n+'
    r'(?P<options>(?:  \1 {0,3}\:[a-zA-Z0-9_-]+\: *[^\n]*\n+)*)'
    r'(?P<text>(?:  \1 {0,3}[^\n]*\n+)*)'
)


class Directive(object):
    @staticmethod
    def parse_text(m):
        text = m.group('text')
        if not text.strip():
            return ''

        leading = len(m.group(1)) + 2
        text = '\n'.join(_parse_text_lines(text, leading)).lstrip('\n') + '\n'
        return text

    @staticmethod
    def parse_options(m):
        text = m.group('options')
        if not text.strip():
            return []

        options = []
        for line in re.split(r'\n+', text):
            line = line.strip()[1:]
            if not line:
                continue
            i = line.find(':')
            k = line[:i]
            v = line[i + 1:].strip()
            options.append((k, v))
        return options

    def register_directive(self, md, name):
        plugin = getattr(md, '_directive', None)
        if not plugin:
            plugin = PluginDirective()
            plugin(md)

        plugin.register_directive(name, self.parse)

    def parse(self, block, m, state):
        raise NotImplementedError()

    def __call__(self, md):
        raise NotImplementedError()


class PluginDirective(object):
    def __init__(self):
        self._directives = {}

    def register_directive(self, name, fn):
        self._directives[name] = fn

    def parse_block_directive(self, block, m, state):
        name = m.group('name')
        method = self._directives.get(name)
        if method:
            return method(block, m, state)

        token = {
            'type': 'block_error',
            'raw': 'Unsupported directive: ' + name,
        }
        return token

    def __call__(self, md):
        md._directive = self
        md.block.register_rule(
            'directive', DIRECTIVE_PATTERN,
            self.parse_block_directive
        )
        md.block.rules.append('directive')


def _parse_text_lines(text, leading):
    spaces = ' ' * leading
    for line in text.splitlines():
        line = line.replace(spaces, '', 1)
        if not line.startswith('    '):
            line = line.strip()
        else:
            line = line.rstrip()
        yield line



# Copyright (c) 2020 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Status rendering logic."""
from tqdm.auto import tqdm

from .scheduling.base import JobStatus
from .util import mistune


def _render_status(
    template,
    template_environment,
    context,
    detailed,
    expand,
    unroll,
    compact,
    output_format,
):
    """Render the status.

    Parameters
    ----------
    template : str
        User provided Jinja2 template file.
    template_environment : :class:`jinja2.Environment`
        Template environment.
    context : dict
        Context that includes all the information for rendering status
        output.
    detailed : bool
        Print a detailed status of each job.
    expand : bool
        Present labels and operations in two separate tables.
    unroll : bool
        Separate columns for jobs and the corresponding operations.
    compact : bool
        Print a compact version of the output.
    output_format : str
        Rendering output format, supports:
        ``'terminal'`` (default), ``'markdown'``, or ``'html'``.

    Returns
    -------
    str
        Status output.

    """
    # Use Jinja2 template for status output
    if template is None:
        if detailed and expand:
            template = "status_expand.jinja"
        elif detailed and not unroll:
            template = "status_stack.jinja"
        elif detailed and compact:
            template = "status_compact.jinja"
        else:
            template = "status.jinja"

    def draw_progress_bar(value, total, escape="", width=40):
        """Visualize progress with a progress bar.

        Parameters
        ----------
        value : int
            The current progress as a fraction of total.
        total : int
            The maximum value that 'value' may obtain.
        width : int
            The character width of the drawn progress bar. (Default value = 40)
        escape : str
            Escape character needed for some formats. (Default value = "")

        Returns
        -------
        str
            Formatted progress bar.

        """
        assert value >= 0 and total > 0
        bar_format = (
            escape
            + f"|{{bar:{width}}}"
            + escape
            + "| {n_fmt}/{total_fmt} ({percentage:<0.2f}%)"
        )
        return tqdm.format_meter(n=value, total=total, elapsed=0, bar_format=bar_format)

    def job_filter(job_op, scheduler_status_code, all_ops):
        """Filter eligible jobs for status print.

        Parameters
        ----------
        job_op : dict
            Operation information for a job.
        scheduler_status_code : dict
            Dictionary information for status code.
        all_ops : bool
            Boolean value indicate if all operations should be displayed.

        Returns
        -------
        bool
            Whether the job is eligible to print.

        """
        return (
            scheduler_status_code[job_op["scheduler_status"]] != "U"
            or job_op["eligible"]
            or all_ops
        )

    def get_operation_status(operation_info, symbols):
        """Determine the status of an operation.

        Parameters
        ----------
        operation_info : dict
            Dictionary containing operation information.
        symbols : dict
            Dictionary containing code for different job statuses.

        Returns
        -------
        str
            The symbol for the job status.

        """
        if operation_info["scheduler_status"] >= JobStatus.active:
            op_status = "running"
        elif operation_info["scheduler_status"] > JobStatus.inactive:
            op_status = "active"
        elif operation_info["completed"]:
            op_status = "completed"
        elif operation_info["eligible"]:
            op_status = "eligible"
        else:
            op_status = "ineligible"

        return symbols[op_status]

    def highlight(string, eligible, pretty):
        """Change font to bold within jinja2 template.

        Parameters
        ----------
        string : str
            The string to be printed.
        eligible : bool
            Boolean value for job eligibility.
        pretty : bool
            Prettify the output.

        Returns
        -------
        str
            The highlighted (bold font) string.

        """
        if eligible and pretty:
            return "**" + string + "**"
        else:
            return string

    template_environment.filters["highlight"] = highlight
    template_environment.filters["draw_progress_bar"] = draw_progress_bar
    template_environment.filters["get_operation_status"] = get_operation_status
    template_environment.filters["job_filter"] = job_filter

    template = template_environment.get_template(template)
    markdown_output = template.render(**context)
    if output_format == "terminal":
        return mistune.terminal(markdown_output)
    elif output_format == "markdown":
        return markdown_output
    elif output_format == "html":
        return mistune.html(markdown_output)
    else:
        raise ValueError(
            "Output format not supported, valid options are "
            "terminal, markdown, or html."
        )



# Copyright (c) 2017 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""The environments module contains a set of provided environment profiles.

These environments are imported by default. This can be disabled by setting
the configuration key 'flow.import_packaged_environments' to 'off' with the
following shell command:

.. code-block:: bash

    signac config --global set flow.import_packaged_environments off

"""
from . import incite, umich, umn, xsede

__all__ = [
    "incite",
    "umich",
    "umn",
    "xsede",
]



import re
from .scanner import ScannerParser, escape_url, unikey

PUNCTUATION = r'''\\!"#$%&'()*+,./:;<=>?@\[\]^`{}|_~-'''
ESCAPE = r'\\[' + PUNCTUATION + ']'
HTML_TAGNAME = r'[A-Za-z][A-Za-z0-9-]*'
HTML_ATTRIBUTES = (
    r'(?:\s+[A-Za-z_:][A-Za-z0-9_.:-]*'
    r'(?:\s*=\s*(?:[^ "\'=<>`]+|\'[^\']*?\'|"[^\"]*?"))?)*'
)
ESCAPE_CHAR = re.compile(r'\\([' + PUNCTUATION + r'])')
LINK_TEXT = r'(?:\[[^\[\]]*\]|\\[\[\]]?|`[^`]*`|[^\[\]\\])*?'
LINK_LABEL = r'(?:[^\\\[\]]|' + ESCAPE + r'){0,1000}'


class InlineParser(ScannerParser):
    ESCAPE = ESCAPE

    #: link or email syntax::
    #:
    #: <https://example.com>
    AUTO_LINK = (
        r'(?<!\\)(?:\\\\)*<([A-Za-z][A-Za-z0-9+.-]{1,31}:'
        r"[^ <>]*?|[A-Za-z0-9.!#$%&'*+/=?^_`{|}~-]+@[A-Za-z0-9]"
        r'(?:[A-Za-z0-9-]{0,61}[A-Za-z0-9])?'
        r'(?:\.[A-Za-z0-9](?:[A-Za-z0-9-]{0,61}[A-Za-z0-9])?)*)>'
    )

    #: link or image syntax::
    #:
    #: [text](/link "title")
    #: ![alt](/src "title")
    STD_LINK = (
        r'!?\[(' + LINK_TEXT + r')\]\(\s*'

        r'(<(?:\\[<>]?|[^\s<>\\])*>|'
        r'(?:\\[()]?|\([^\s\x00-\x1f\\]*\)|[^\s\x00-\x1f()\\])*?)'

        r'(?:\s+('
        r'''"(?:\\"?|[^"\\])*"|'(?:\\'?|[^'\\])*'|\((?:\\\)?|[^)\\])*\)'''
        r'))?\s*\)'
    )

    #: Get link from references. References are defined in DEF_LINK in blocks.
    #: The syntax looks like::
    #:
    #:    [an example][id]
    #:
    #:    [id]: https://example.com "optional title"
    REF_LINK = (
        r'!?\[(' + LINK_TEXT + r')\]'
        r'\[(' + LINK_LABEL + r')\]'
    )

    #: Simple form of reference link::
    #:
    #:    [an example]
    #:
    #:    [an example]: https://example.com "optional title"
    REF_LINK2 = r'!?\[(' + LINK_LABEL + r')\]'

    #: emphasis and strong * or _::
    #:
    #:    *emphasis*  **strong**
    #:    _emphasis_  __strong__
    ASTERISK_EMPHASIS = (
        r'(\*{1,2})(?=[^\s*])('
        r'(?:\\[\\*]|[^*])*'
        r'(?:' + ESCAPE + r'|[^\s*]))\1'
    )
    UNDERSCORE_EMPHASIS = (
        r'\b(_{1,2})(?=[^\s_])([\s\S]*'
        r'(?:' + ESCAPE + r'|[^\s_]))\1'
        r'(?!_|[^\s' + PUNCTUATION + r'])\b'
    )

    #: codespan with `::
    #:
    #:    `code`
    CODESPAN = (
        r'(?<!\\|`)(?:\\\\)*(`+)(?!`)([\s\S]+?)(?<!`)\1(?!`)'
    )

    #: linebreak leaves two spaces at the end of line
    LINEBREAK = r'(?:\\| {2,})\n(?!\s*$)'

    INLINE_HTML = (
        r'(?<!\\)<' + HTML_TAGNAME + HTML_ATTRIBUTES + r'\s*/?>|'  # open tag
        r'(?<!\\)</' + HTML_TAGNAME + r'\s*>|'  # close tag
        r'(?<!\\)<!--(?!>|->)(?:(?!--)[\s\S])+?(?<!-)-->|'  # comment
        r'(?<!\\)<\?[\s\S]+?\?>|'
        r'(?<!\\)<![A-Z][\s\S]+?>|'  # doctype
        r'(?<!\\)<!\[CDATA[\s\S]+?\]\]>'  # cdata
    )

    RULE_NAMES = (
        'escape', 'inline_html', 'auto_link',
        'std_link', 'ref_link', 'ref_link2',
        'asterisk_emphasis', 'underscore_emphasis',
        'codespan', 'linebreak',
    )

    def __init__(self, renderer):
        super(InlineParser, self).__init__()
        self.renderer = renderer
        rules = list(self.RULE_NAMES)
        rules.remove('ref_link')
        rules.remove('ref_link2')
        self.ref_link_rules = rules

    def parse_escape(self, m, state):
        text = m.group(0)[1:]
        return 'text', text

    def parse_auto_link(self, m, state):
        if state.get('_in_link'):
            return 'text', m.group(0)

        text = m.group(1)
        if '@' in text and not text.lower().startswith('mailto:'):
            link = 'mailto:' + text
        else:
            link = text
        return 'link', escape_url(link), text

    def parse_std_link(self, m, state):
        line = m.group(0)
        text = m.group(1)
        link = ESCAPE_CHAR.sub(r'\1', m.group(2))
        if link.startswith('<') and link.endswith('>'):
            link = link[1:-1]

        title = m.group(3)
        if title:
            title = ESCAPE_CHAR.sub(r'\1', title[1:-1])

        if line[0] == '!':
            return 'image', link, text, title

        return self.tokenize_link(line, link, text, title, state)

    def parse_ref_link(self, m, state):
        line = m.group(0)
        text = m.group(1)
        key = unikey(m.group(2) or text)
        def_links = state.get('def_links')
        if not def_links or key not in def_links:
            return list(self._scan(line, state, self.ref_link_rules))

        link, title = def_links.get(key)
        link = ESCAPE_CHAR.sub(r'\1', link)
        if title:
            title = ESCAPE_CHAR.sub(r'\1', title)

        if line[0] == '!':
            return 'image', link, text, title

        return self.tokenize_link(line, link, text, title, state)

    def parse_ref_link2(self, m, state):
        return self.parse_ref_link(m, state)

    def tokenize_link(self, line, link, text, title, state):
        if state.get('_in_link'):
            return 'text', line
        state['_in_link'] = True
        text = self.render(text, state)
        state['_in_link'] = False
        return 'link', escape_url(link), text, title

    def parse_asterisk_emphasis(self, m, state):
        return self.tokenize_emphasis(m, state)

    def parse_underscore_emphasis(self, m, state):
        return self.tokenize_emphasis(m, state)

    def tokenize_emphasis(self, m, state):
        marker = m.group(1)
        text = m.group(2)
        if len(marker) == 1:
            return 'emphasis', self.render(text, state)
        return 'strong', self.render(text, state)

    def parse_codespan(self, m, state):
        code = re.sub(r'[ \n]+', ' ', m.group(2).strip())
        return 'codespan', code

    def parse_linebreak(self, m, state):
        return 'linebreak',

    def parse_inline_html(self, m, state):
        html = m.group(0)
        return 'inline_html', html

    def parse_text(self, text, state):
        return 'text', text

    def parse(self, s, state, rules=None):
        if rules is None:
            rules = self.rules

        tokens = (
            self.renderer._get_method(t[0])(*t[1:])
            for t in self._scan(s, state, rules)
        )
        return tokens

    def render(self, s, state, rules=None):
        tokens = self.parse(s, state, rules)
        if self.renderer.IS_TREE:
            return list(tokens)
        return ''.join(tokens)

    def __call__(self, s, state):
        return self.render(s, state)



from .scanner import escape, escape_html
from .plugins.tabulate import tabulate


class BaseRenderer(object):
    NAME = 'base'
    IS_TREE = False

    def __init__(self):
        self._methods = {}

    def register(self, name, method):
        self._methods[name] = method

    def _get_method(self, name):
        try:
            return object.__getattribute__(self, name)
        except AttributeError:
            method = self._methods.get(name)
            if not method:
                raise AttributeError('No renderer "{!r}"'.format(name))
            return method


class AstRenderer(BaseRenderer):
    NAME = 'ast'
    IS_TREE = True

    def text(self, text):
        return {'type': 'text', 'text': text}

    def link(self, link, children=None, title=None):
        return {'type': 'link', 'link': link, 'children': children, 'title': title}

    def image(self, src, alt="", title=None):
        return {'type': 'image', 'src': src, 'alt': alt, 'title': title}

    def codespan(self, text):
        return {'type': 'codespan', 'text': text}

    def linebreak(self):
        return {'type': 'linebreak'}

    def inline_html(self, html):
        return {'type': 'inline_html', 'text': html}

    def heading(self, children, level):
        return {'type': 'heading', 'children': children, 'level': level}

    def newline(self):
        return {'type': 'newline'}

    def thematic_break(self):
        return {'type': 'thematic_break'}

    def block_code(self, children, info=None):
        return {
            'type': 'block_code',
            'text': children,
            'info': info
        }

    def block_html(self, children):
        return {'type': 'block_html', 'text': children}

    def list(self, children, ordered, level, start=None):
        token = {
            'type': 'list',
            'children': children,
            'ordered': ordered,
            'level': level,
        }
        if start is not None:
            token['start'] = start
        return token

    def list_item(self, children, level):
        return {'type': 'list_item', 'children': children, 'level': level}

    def _create_default_method(self, name):
        def __ast(children):
            return {'type': name, 'children': children}
        return __ast

    def _get_method(self, name):
        try:
            return super(AstRenderer, self)._get_method(name)
        except AttributeError:
            return self._create_default_method(name)


class HTMLRenderer(BaseRenderer):
    NAME = 'html'
    IS_TREE = False
    HARMFUL_PROTOCOLS = {
        'javascript:',
        'vbscript:',
        'data:',
    }

    def __init__(self, escape=True, allow_harmful_protocols=None):
        super(HTMLRenderer, self).__init__()
        self._escape = escape
        self._allow_harmful_protocols = allow_harmful_protocols

    def _safe_url(self, url):
        if self._allow_harmful_protocols is None:
            schemes = self.HARMFUL_PROTOCOLS
        elif self._allow_harmful_protocols is True:
            schemes = None
        else:
            allowed = set(self._allow_harmful_protocols)
            schemes = self.HARMFUL_PROTOCOLS - allowed

        if schemes:
            for s in schemes:
                if url.startswith(s):
                    url = '#harmful-link'
                    break
        return url

    def text(self, text):
        return escape_html(text)

    def link(self, link, text=None, title=None):
        if text is None:
            text = link

        s = '<a href="' + self._safe_url(link) + '"'
        if title:
            s += ' title="' + escape_html(title) + '"'
        return s + '>' + (text or link) + '</a>'

    def image(self, src, alt="", title=None):
        s = '<img src="' + src + '" alt="' + alt + '"'
        if title:
            s += ' title="' + escape_html(title) + '"'
        return s + ' />'

    def emphasis(self, text):
        return '<em>' + text + '</em>'

    def strong(self, text):
        return '<strong>' + text + '</strong>'

    def codespan(self, text):
        return '<code>' + escape(text) + '</code>'

    def linebreak(self):
        return '<br />\n'

    def inline_html(self, html):
        if self._escape:
            return escape(html)
        return html

    def paragraph(self, text):
        return '<p>' + text + '</p>\n'

    def heading(self, text, level):
        tag = 'h' + str(level)
        return '<' + tag + '>' + text + '</' + tag + '>\n'

    def newline(self):
        return ''

    def thematic_break(self):
        return '<hr />\n'

    def block_text(self, text):
        return text

    def block_code(self, code, info=None):
        html = '<pre><code'
        if info:
            lang = info.strip().split(None, 1)[0]
            lang = escape_html(lang)
            html += ' class="language-' + lang + '"'
        return html + '>' + escape(code) + '</code></pre>\n'

    def block_quote(self, text):
        return '<blockquote>\n' + text + '</blockquote>\n'

    def block_html(self, html):
        if not self._escape:
            return html + '\n'
        return '<p>' + escape(html) + '</p>\n'

    def block_error(self, html):
        return '<div class="error">' + html + '</div>\n'

    def list(self, text, ordered, level, start=None):
        if ordered:
            html = '<ol'
            if start is not None:
                html += ' start="' + str(start) + '"'
            return html + '>\n' + text + '</ol>\n'
        return '<ul>\n' + text + '</ul>\n'

    def list_item(self, text, level):
        return '<li>' + text + '</li>\n'


class TextRenderer(BaseRenderer):
    NAME = 'text'
    IS_TREE = False

    def __init__(self):
        super(TextRenderer, self).__init__()

    def text(self, text):
        return text

    def heading(self, text, level):
        return '#' * level + ' ' + text.strip() + '\n\n'

    def list(self, text, ordered, level, start=None):
        assert not ordered
        return text

    def list_item(self, text, level):
        return ' ' * (level-1) + '- ' + text + '\n'

    def block_text(self, text):
        return text

    def thematic_break(self):
        return '\n---\n\n'

    def paragraph(self, text):
        return text.strip('\n') + '\n\n'

    def link(self, link, text=None, title=None):
        return '[{text}]({link})'.format(text=text, link=link)

    def block_quote(self, text):
        return '\n'.join(['> ' + line for line in text.strip().splitlines()]) + '\n'

    def newline(self):
        return '\n'

    def block_code(self, code, info=None):
        return '```{}\n'.format(info if info else '') + code.strip('\n') + "\n```\n\n"

    def codespan(self, text):
        return '`{text}`'.format(text=text)

    def strong(self, text):
        return '**' + text + '**'

    def linebreak(self):
        return '\n'

    def table(self, text):
        return tabulate(text['rows'], headers=text['headers']) + self.linebreak() * 2


class TerminalRenderer(TextRenderer):
    NAME = 'terminal'
    IS_TREE = False

    def __init__(self):
        super(TerminalRenderer, self).__init__()

    def block_code(self, code, info=None):
        return '\033[48;100m{}\n'.format(info if info else '') + code.strip('\n') + '\n\033[0m\n\n'

    def codespan(self, text):
        return '\033[48;100m{text}\033[0m'.format(text=text)

    def strong(self, text):
        return '\033[1m' + text + '\033[0m'

    def modify_strong(self, text):
        if text.strip('*') == text:
            return text
        else:
            return self.strong(text.strip('*'))

    def table(self, text):
        rows = [[self.modify_strong(cell) for cell in row] for row in text['rows']]
        return tabulate(rows, headers=text['headers']) + self.linebreak() * 2
